/* 
Keyless electronic combination lock usies optical finger navigation sensor

Use with Arduino IDE version 1.0 or later (NOT compatible with pre-release versions)

IMPORTANT! If the OFN sensor does not operate reliably, or at all, you may need to
add 1.8k to 3.3k pullup resistors between pins A4 & 5V, and A5 & 5V.

Uses a Parallax Optical Finger Navigation (OFN) sensor as a keyless combination
lock. This sensor detects finger gesters in X and Y directions, and is interpreted
as up, down, left, and right movement. The sensor is mounted on the side of a wooden
cigar box that serves as an electronic lock box. An R/C servo motor provides the
locking action against a metal retainer bar mounted on the top lid of the box.

A preset combination is stored in the sketch (see the combination[] array), 
and if the same sequence is entered on the OFN sensor a servo motor is activated 
to unlock the box.

The more steps in the combination, the better the security of the lock, 
but the easier it is to make a mistake. If you find the combination isn't 
working reliably, slow down, and be sure your finger moves in straight 
up/down and left/right directions. 

To change the preset combination modify the combination[] array, and change
MAX_DIRS to the same or fewer elements in the array. Example:

char combination[] = {'l','r','l','d','u'};
#define MAX_DIRS      5

specified a sequence of 5 finger gestures: left, right, left, down, and up. Note
that each element of the array is enclosed in single quotes, and separated with
a comma.

Variables you may wish to change:

MIN_VAL - Determines the smallest delta value from the OFN that triggers a movement
RELOCK_VAL - Specifies the time delay to relock the box
ERVO_LOCK - Specifies the servo degree angle for locking the box
SERVO_UNLOCK - Specifies the servo degree angle for unlocking the box

Gordon McComb - www.robotoid.com
*/

#include <Servo.h>              // Library for servo motor
#include <Wire.h>               // Library for I2C comm with OFN module
#include "SimpleTimer.h"        // Library for background timer (for relocking)
#include "pitches.h"            // Definitions for musical notes

// Configure combination lock
char combination[] = {'l','r','l'};  // Preset combination (left-right-left)
#define MAX_DIRS      3              // Number of directions in combination

// Delays and minimum values to register as movement (helps to prevent mis-reads)
#define CMD_DELAY     150       // Delay between successive commands
#define REP_DELAY     25        // Minimum repetition delay
#define MIN_VAL       12        // Minimum value before recognizing as movement
#define RELOCK_VAL    7000      // Time to relock (in milliseconds)

// OFN pin designations
// MOT = D2; implemented as hardware interrupt 0
#define SW            3
#define RST           4
#define SHD           5
#define GND           6
#define PWR           7
#define ofnID         0x66 >> 1  // 7-bit I2C address of OFN

// Servo definitions
#define SERVO_PIN     8          // Connected to pin 8
#define SERVO_LOCK    90         // Lock position
#define SERVO_UNLOCK  170        // Unlock position

// Speaker definitions
#define SPKR          11
#define SPKR_GND      12

// LED definition
#define LED_PIN       13

volatile boolean motionFlag = false;
char deltaX = 0;
char deltaY = 0;
byte tempVal = 0xFF;

// Structure for holding combination values
typedef struct{
  char lastCmd;
  char thisCmd;
  byte cmdCount;
  unsigned long lastTime;
  char ofnDirs[MAX_DIRS];
  byte countDirs;
}OfnCmd;
OfnCmd cmd = {' ', ' ', 0, millis(),{' '}, 0};

SimpleTimer timer;
Servo lockServo; 

void setup() {
  
  // Set pin for speaker ground
  pinMode(SPKR_GND, OUTPUT);
  digitalWrite(SPKR_GND, LOW);
  
  // Set up power pins for OFN module
  pinMode(GND, OUTPUT);
  pinMode(PWR, OUTPUT);
  digitalWrite(GND, LOW);
  digitalWrite(PWR, HIGH);

  // General setup
  Serial.begin(9600);
  Wire.begin();
  pinMode(LED_PIN, OUTPUT);
  delay(100);
  
  // Reset OFN
  pinMode(RST, OUTPUT);
  digitalWrite(RST, LOW);
  pinMode(RST, INPUT);
  
  // Check if OFN is available/online
  if(getID() != 0x83)
    Serial.println("Device not found");
  else
    Serial.println("Device ready");
  
  configDevice();
  attachInterrupt(0, motionDetect, FALLING);
  
  // Failsafe - temporarily unlocks box
  //  To trigger depress OFN button during Arduino power-on or reset
  //  Remove or comment out this code if you don't want this behavior
  if (digitalRead(SW) == 0) {
    unlock_the_servo();
    delay(1500);         // 1.5 second delay
  }
  lock_the_servo();    
 
}

void loop() {
 
  // If motion interrupt has occurred since last loop
  if (motionFlag) {
    motionFlag = false;
    
    deltaX = getX();
    deltaY = getY();

    if(deltaY > MIN_VAL)
      cmd.thisCmd = 'd';
    
    if(deltaY < 0 && abs(deltaY) > MIN_VAL)
      cmd.thisCmd = 'u';
      
    if(deltaX > MIN_VAL)
      cmd.thisCmd = 'r';
    
    if(deltaX < 0 && abs(deltaX) > MIN_VAL)
     cmd.thisCmd = 'l';
    
    // Prevent "double swiping"
    if (millis() - cmd.lastTime < CMD_DELAY) {
      if(cmd.thisCmd == cmd.lastCmd)
        cmd.cmdCount++;
    } else {
      cmd.cmdCount = 0;
      cmd.thisCmd = ' ';
    }
   
    // Store command value
    cmd.lastTime = millis();
    if(cmd.thisCmd != ' ') {
      if(cmd.lastCmd != cmd.thisCmd) {
        if(cmd.countDirs < MAX_DIRS) {
          cmd.ofnDirs[cmd.countDirs] = cmd.thisCmd;
          Serial.print(cmd.countDirs, DEC);
          Serial.print("\t");
          Serial.println(cmd.thisCmd);
          cmd.countDirs++;
        }
      }
    }
  
    cmd.lastCmd = cmd.thisCmd; 
    clearMotion();
    delay(REP_DELAY);
    
   }
  
   // If OFN button is depressed validate combination
   if (digitalRead(SW) == 0) {
     if (compareCombination()) {        // Combination good; unlock servo
       unlock_the_servo();
       timer.setTimeout(RELOCK_VAL, lock_the_servo);
     } else {
       Serial.println("Combination failed");
       int tones[] = {NOTE_E2};
       makeTone(tones, sizeof(tones)/sizeof(int));
     }
     cmd.countDirs = 0;                // Reset entered combination
     delay(200);
   }
   
   timer.run();                        // Update simple timer on each loop
  
}

// Unlock servo
void unlock_the_servo() {
  Serial.println("Servo in unlock position");
  digitalWrite(LED_PIN, HIGH);
  int tones[] = {NOTE_C4, NOTE_B3, NOTE_C4};
  makeTone(tones, sizeof(tones)/sizeof(int));
  lockServo.attach(SERVO_PIN);
  lockServo.write(SERVO_UNLOCK);
  delay(500);
  lockServo.detach();
}

// Lock servo
void lock_the_servo() {
  Serial.println("Servo in lock position");
  digitalWrite(LED_PIN, LOW);  
  int tones[] = {NOTE_B3, NOTE_B3};
  makeTone(tones, sizeof(tones)/sizeof(int));
  lockServo.attach(SERVO_PIN);
  lockServo.write(SERVO_LOCK);
  delay(500);
  lockServo.detach();
}

// Validate combination
//  Compare stored combination with entered combination
boolean compareCombination() { 
  for (int i = 0; i < MAX_DIRS; i++) {
    if (combination[i] != cmd.ofnDirs[i])
      return(false);
  }
  return(true);
}

// Interrupt handler if motion is detected (pin D2 is triggered by MOT pin of OFN module)
void motionDetect() {
  motionFlag = true;
}

// Iterate notes of tune
void makeTone(int tones[], int length) {
  int toneDuration = 1000/4;
  for (int thisNote = 0; thisNote < length; thisNote++) {
    tone(SPKR, tones[thisNote], toneDuration);  // Note duration
    delay(toneDuration * 1.30);                 // Pause between
    noTone(SPKR);                               // Stop tone
  }
  return;
}

//----- OFN functions -----
// Clear X/Y motion register
void clearMotion() {
  Wire.beginTransmission(ofnID);
  Wire.write(byte(0x02));
  Wire.write(byte(0x00));
  Wire.endTransmission();
}

// Get X movement
byte getX() {
  Wire.beginTransmission(ofnID);
  Wire.write(byte(0x03));
  Wire.endTransmission();
  Wire.requestFrom(ofnID, 1);
  return(Wire.read());
}

// Get Y movement
byte getY() {
  Wire.beginTransmission(ofnID);
  Wire.write(byte(0x04));
  Wire.endTransmission();
  Wire.requestFrom(ofnID, 1);
  return(Wire.read());  
}

// Configure OFN module
void configDevice() {
  Wire.beginTransmission(ofnID);
  Wire.write(byte(0x77));
  Wire.write(B00100001);                   // Invert X, rotate 0 degrees
  Wire.endTransmission();
}

// Get OFN ID (validates if sesnor is online)
byte getID() {
  Wire.beginTransmission(ofnID);
  Wire.write(byte(0x00)); 
  Wire.endTransmission();
  Wire.requestFrom(ofnID, 1); 
  return(Wire.read());
}
